{ **********************************************************************
  *                            Unit FMATH.PAS                          *
  *                              Version 1.6                           *
  *                      (c) J. Debord, August 1997                    *
  **********************************************************************
  This unit implements several mathematical functions in Turbo Pascal.

  To compile this unit (and the other units or programs which use it),
  use the following compiler settings :

  1) If you have a math coprocessor, set "Options/Compiler/Numeric
     processing" to 8087/80287 ($N+ mode). If you have no coprocessor,
     you may wish to use the Emulation ($E+) mode.

  2) If you have Turbo Pascal 6.0 (or higher) and a 387 (or higher)
     coprocessor, define the symbol CPU387 in "Options/Compiler/
     Conditional defines".

     This will ensure the inclusion of the file MATH387.INC which
     contains assembler routines. These routines make direct calls
     to the coprocessor functions, which run faster than their
     Turbo Pascal analogs.

  NB: After any modification of the "Conditional defines", be sure
      to recompile all the TPMATH units and programs (i.e. "Build",
      not "Make").
  **********************************************************************
  ERROR HANDLING : The global variable MathErr contains the error code
  returned by the last function evaluation. It may be tested immediately
  after a function call:

    Y := Ln(X);
    if MathErr = FN_OK then ...

  The possible error codes, and the default values attributed to the
  function, are the following:

  Error code   Value  Significance            Function default value
  ------------------------------------------------------------------
  FN_OK          0    No error
  DOMAIN        -1    Argument domain error   0
  SING          -2    Function singularity    +/- MAXNUM
  OVERFLOW      -3    Overflow range error    MAXNUM
  UNDERFLOW     -4    Underflow range error   0

  where MAXNUM is a constant defining the highest number which may be
  represented within the chosen floating point type.
  **********************************************************************
  References:
  1) Cephes math library by Stephen L. Moshier : Gamma, LnGamma.
  2) 'Numerical Recipes' by Press et al. : other special functions and
     Gaussian random number generator.
  3) RANLAF.C by L.A. Froistad : Marsaglia random number generator.
  4) MATHLIB2 by J.D. Gayrard : Assembler routines for the 387
  ********************************************************************** }

unit FMath;

interface

{ ----------------------------------------------------------------------
  Set floating point type and constants according to compiler directives
  ---------------------------------------------------------------------- }

{$IFOPT N+}
type
  Float = Extended;
const
  MACHEP = 1.08420217248550444E-19;   { Floating point precision: 2^(-63) }
  MAXNUM = 1.18973149535723103E+4932; { Highest floating point number: 2^16384 }
  MINNUM = 3.36210314311209558E-4932; { Lowest floating point number: 2^(-16382) }
  MAXLOG = 11356.5234062941439;       { Highest argument for Exp = Ln(MAXNUM) }
  MINLOG = -11355.137111933024;       { Lowest argument for Exp = Ln(MINNUM) }
  MAXFAC = 1754;                      { Highest argument for Factorial }
  MAXGAM = 1755.455;                  { Highest argument for Gamma }
  MAXLGM = 1.04848146839019521E+4928; { Highest argument for LnGamma }
{$ELSE}
type
  Float = Real;
const
  MACHEP = 1.818989404E-12;  { 2^(-39) }
  MAXNUM = 4.253529586E+37;  { 2^126 }
  MINNUM = 2.350988703E-38;  { 2^(-125) }
  MAXLOG = 8.664339757E+01;
  MINLOG = - 4.253529586E+01;
  MAXFAC = 33;
  MAXGAM = 34.64809785;
  MAXLGM = 1.038324114E+36;
{$ENDIF}

{ *************************** Error codes ****************************** }

const
  FN_OK = 0;        { No error }
  DOMAIN = - 1;     { Argument domain error }
  SING = - 2;       { Function singularity }
  OVERFLOW = - 3;   { Overflow range error }
  UNDERFLOW = - 4;  { Underflow range error }

{ *********************** Mathematical constants *********************** }

const
  LN2 = 0.69314718055994530942;        { Ln(2) }
  LN10 = 2.30258509299404568402;       { Ln(10) }
  LNPI = 1.14472988584940017414;       { Ln(Pi) }
  INVLN2 = 1.44269504088896340736;     { 1/Ln(2) }
  INVLN10 = 0.43429448190325182765;    { 1/Ln(10) }
  TWOPI = 6.28318530717958647693;      { 2*Pi }
  PIDIV2 = 1.57079632679489661923;     { Pi/2 }
  SQRTPI = 1.77245385090551602730;     { Sqrt(Pi) }
  SQRT2PI = 2.50662827463100050242;    { Sqrt(2*Pi) }
  LNSQRT2PI = 0.91893853320467274178;  { Ln(Sqrt(2*Pi)) }
  SQRT2 = 1.41421356237309504880;      { Sqrt(2) }
  SQRT2DIV2 = 0.70710678118654752440;  { Sqrt(2)/2 }

{ ************************** Global variable *************************** }

var
  MathErr : Integer;  { Error code of the last function evaluation }

{ ************ Minimum, maximum and exchange of two numbers ************ }

function FMin(X, Y : Float) : Float;      { Minimum of 2 reals }
function FMax(X, Y : Float) : Float;      { Maximum of 2 reals }
function IMin(X, Y : Integer) : Integer;  { Minimum of 2 integers }
function IMax(X, Y : Integer) : Integer;  { Maximum of 2 integers }
procedure Swap(var X, Y : Float);         { Exchange 2 reals }
procedure ISwap(var X, Y : Integer);      { Exchange 2 integers }

{ ************** Sign, logarithms, exponentials and power ************** }

function Sgn(X : Float) : Integer;                { Sign }
function Exp(X : Float) : Float;                  { Exponential }
function Exp2(X : Float) : Float;                 { 2^X }
function Exp10(X : Float) : Float;                { 10^X }
function Ln(X : Float) : Float;                   { Natural log }
function Log2(X : Float) : Float;                 { Log, base 2 }
function Log10(X : Float) : Float;                { Decimal log }
function Log(X, A : Float) : Float;               { Log, base A }
function IntPow(X : Float; N : Integer) : Float;  { X^N }
function Pow(X, Y : Float) : Float;               { X^Y, X >= 0 }
function Pythag(X, Y : Float) : Float;            { Sqrt(X^2 + Y^2) }

{ ********** Trigonometric and inverse trigonometric functions ********* }

function FixAngle(Theta : Float) : Float;  { Set Theta in -Pi..Pi }
function Sin(X : Float) : Float;           { Sinus }
function Cos(X : Float) : Float;           { Cosinus }
function Tan(X : Float) : Float;           { Tangent }
function ArcSin(X : Float) : Float;        { Arc sinus }
function ArcCos(X : Float) : Float;        { Arc cosinus }
function ArcTan(X : Float) : Float;        { Arc tangent }
function ArcTan2(X, Y : Float) : Float;    { Angle (Ox, OM) with M(X,Y) }

procedure SinCos(X : Float; var SinX, CosX : Float);  { Sin & Cos }

{ ************* Hyperbolic and inverse hyperbolic functions ************ }

function Sinh(X : Float) : Float;     { Hyperbolic sine }
function Cosh(X : Float) : Float;     { Hyperbolic cosine }
function Tanh(X : Float) : Float;     { Hyperbolic tangent }
function ArcSinh(X : Float) : Float;  { Inverse hyperbolic sine }
function ArcCosh(X : Float) : Float;  { Inverse hyperbolic cosine }
function ArcTanh(X : Float) : Float;  { Inverse hyperbolic tangent }

procedure SinhCosh(X : Float; var SinhX, CoshX : Float);  { Sinh & Cosh }

{ ************************** Special functions ************************* }

function Fact(N : Integer) : Float;         { Factorial }
function Binomial(N, K : Integer) : Float;  { Binomial coef. C(N,K) }
function Gamma(X : Float) : Float;          { Gamma function }
function SgnGamma(X : Float) : Integer;     { Sign of Gamma function }
function LnGamma(X : Float) : Float;        { Log(|Gamma(X)|) }
function IGamma(A, X : Float) : Float;      { Incomplete Gamma function }
function JGamma(A, X : Float) : Float;      { Complement of IGamma }
function Beta(X, Y : Float) : Float;        { Beta function }
function IBeta(A, B, X : Float) : Float;    { Incomplete Beta function }
function Erf(X : Float) : Float;            { Error function }
function Erfc(X : Float) : Float;           { Complement of Erf }

{ ********************** Probability distributions ********************* }

function PNorm(X : Float) : Float;
{ ----------------------------------------------------------------------
  Probability of standard normal (U) : PNorm(X) = Prob(|U| >= |X|)
  ---------------------------------------------------------------------- }

function PStudent(Nu : Integer; X : Float) : Float;
{ ----------------------------------------------------------------------
  Probability of Student's t : PStudent(Nu, X) = Prob(|t| >= |X|)
                               Nu = degrees of freedom
  ---------------------------------------------------------------------- }

function PKhi2(Nu : Integer; X : Float) : Float;
{ ----------------------------------------------------------------------
  Probability of Pearson's Khi-2 : PKhi2(Nu, X) = Prob(Khi2 >= X)
                                   Nu = degrees of freedom
  ---------------------------------------------------------------------- }

function PSnedecor(Nu1, Nu2 : Integer; X : Float) : Float;
{ ----------------------------------------------------------------------
  Probability of Snedecor's F : PSnedecor(Nu1, Nu2, X) = Prob(F >= X)
                                Nu1, Nu2 = degrees of freedom
                                           (numerator, denominator)
  ---------------------------------------------------------------------- }

{ *************************** Random numbers *************************** }

procedure RMarIn(Seed1, Seed2 : Integer);
{ ----------------------------------------------------------------------
  Initializes the Marsaglia random number generator.
  Input parameters : 0 <= Seed1 <= 31328, 0 <= Seed2 <= 30081
  The default initialization corresponds to RMarIn(1802, 9373).
  ---------------------------------------------------------------------- }

function IRanMar : LongInt;
{ ----------------------------------------------------------------------
  Returns a random number in [0, 16777216[
  ---------------------------------------------------------------------- }

function RanMar : Float;
{ ----------------------------------------------------------------------
  Returns a random number in ]0, 1[
  ---------------------------------------------------------------------- }

function RanGauss(Mu, Sigma : Float) : Float;
{ ----------------------------------------------------------------------
  Returns a random number from a Gaussian distribution of mean Mu and
  standard deviation Sigma
  ---------------------------------------------------------------------- }

{ ********************************************************************** }

implementation

{ Constants and variables used to build a table of factorials }
const
  NFACT = 33;
var
  FactArray : array[0..NFACT] of Float;
  I : Integer;

{ Constants and variables used by the random number generator }
const
  LASTDIM = 97;
  FIRSTDIM = 65;
  NO_NUMBERS = 16777216;  { 2^24 }
  CM = 16777213;          { NO_NUMBERS - 3 }
  CD = 7654321;
var
  X : array[1..LASTDIM] of LongInt;
  J_ptr, K_ptr, C : LongInt;

{$IFDEF CPU387}
  {$I MATH387.INC}
{$ENDIF}

  function FMin(X, Y : Float) : Float;
  begin
    if X <= Y then
      FMin := X
    else
      FMin := Y;
  end;

  function FMax(X, Y : Float) : Float;
  begin
    if X >= Y then
      FMax := X
    else
      FMax := Y;
  end;

  function IMin(X, Y : Integer) : Integer;
  begin
    if X <= Y then
      IMin := X
    else
      IMin := Y;
  end;

  function IMax(X, Y : Integer) : Integer;
  begin
    if X >= Y then
      IMax := X
    else
      IMax := Y;
  end;

  procedure Swap(var X, Y : Float);
  var
    Temp : Float;
  begin
    Temp := X;
    X := Y;
    Y := Temp;
  end;

  procedure ISwap(var X, Y : Integer);
  var
    Temp : Integer;
  begin
    Temp := X;
    X := Y;
    Y := Temp;
  end;

  function Sgn(X : Float) : Integer;
  begin
    if X > 0.0 then
      Sgn := 1
    else if X = 0.0 then
      Sgn := 0
    else
      Sgn := - 1;
  end;

  function DefaultVal(ErrCode : Integer) : Float;
  { Sets the global variable MathErr and the function default value
    according to the error code }
  begin
    MathErr := ErrCode;
    case ErrCode of
      DOMAIN : DefaultVal := 0.0;
      SING : DefaultVal := MAXNUM;
      OVERFLOW : DefaultVal := MAXNUM;
      UNDERFLOW : DefaultVal := 0.0;
    end;
  end;

  function Exp(X : Float) : Float;
  begin
    MathErr := FN_OK;
    if X < MINLOG then
      Exp := DefaultVal(UNDERFLOW)
    else if X > MAXLOG then
      Exp := DefaultVal(OVERFLOW)
    else
    {$IFDEF CPU387}
      Exp := fexp(X);
    {$ELSE}
      Exp := System.Exp(X);
    {$ENDIF}
  end;

  function Exp2(X : Float) : Float;
  var
    XLn2 : Float;
  begin
    MathErr := FN_OK;
    XLn2 := X * LN2;
    if XLn2 < MINLOG then
      Exp2 := DefaultVal(UNDERFLOW)
    else if XLn2 > MAXLOG then
      Exp2 := DefaultVal(OVERFLOW)
    else
    {$IFDEF CPU387}
      Exp2 := fexp2(X);
    {$ELSE}
      Exp2 := System.Exp(XLn2);
    {$ENDIF}
  end;

  function Exp10(X : Float) : Float;
  var
    XLn10 : Float;
  begin
    MathErr := FN_OK;
    XLn10 := X * LN10;
    if XLn10 < MINLOG then
      Exp10 := DefaultVal(UNDERFLOW)
    else if XLn10 > MAXLOG then
      Exp10 := DefaultVal(OVERFLOW)
    else
    {$IFDEF CPU387}
      Exp10 := ften_to(X);
    {$ELSE}
      Exp10 := System.Exp(XLn10);
    {$ENDIF}
  end;

  function Ln(X : Float) : Float;
  begin
    MathErr := FN_OK;
    if X < 0.0 then
      Ln := DefaultVal(DOMAIN)
    else if X = 0.0 then
      Ln := DefaultVal(SING)
    else
    {$IFDEF CPU387}
      Ln := fln(X);
    {$ELSE}
      Ln := System.Ln(X);
    {$ENDIF}
  end;

  function Log10(X : Float) : Float;
  begin
    MathErr := FN_OK;
    if X < 0.0 then
      Log10 := DefaultVal(DOMAIN)
    else if X = 0.0 then
      Log10 := DefaultVal(SING)
    else
    {$IFDEF CPU387}
      Log10 := flog(X);
    {$ELSE}
      Log10 := System.Ln(X) * INVLN10;
    {$ENDIF}
  end;

  function Log2(X : Float) : Float;
  begin
    MathErr := FN_OK;
    if X < 0.0 then
      Log2 := DefaultVal(DOMAIN)
    else if X = 0.0 then
      Log2 := DefaultVal(SING)
    else
    {$IFDEF CPU387}
      Log2 := flog2(X)
    {$ELSE}
      Log2 := System.Ln(X) * INVLN2;
    {$ENDIF}
  end;

  function Log(X, A : Float) : Float;
  begin
    MathErr := FN_OK;
    if (X < 0.0) or (A <= 0.0) or (A = 1.0) then
      Log := DefaultVal(DOMAIN)
    else if X = 0.0 then
      Log := Sgn(1.0 - A) * DefaultVal(SING)
    else
    {$IFDEF CPU387}
      Log := fln(X) / fln(A);
    {$ELSE}
      Log := System.Ln(X) / System.Ln(A);
    {$ENDIF}
  end;

  function IntPow(X : Float; N : Integer) : Float;
  { Computes X^N by repeated multiplications }
  var
    M : Integer;
    T : Float;
  begin
    MathErr := FN_OK;

    if X = 0.0 then
      begin
        if N = 0 then       { 0^0 = lim  x^x = 1 }
          IntPow := 1.0     {       x->0         }
        else if N > 0 then
          IntPow := 0.0     { 0^N = 0 }
        else
          IntPow := DefaultVal(SING);
        Exit;
      end;

    if N = 0 then
      begin
        IntPow := 1.0;
        Exit;
      end;

    { Legendre's algorithm for minimizing the number of multiplications }
    T := 1.0;
    M := Abs(N);
    repeat
      if Odd(M) then
        begin
          Dec(M);
          T := T * X;
        end
      else
        begin
          M := M div 2;
          X := Sqr(X);
        end;
    until M = 0;

    if N > 0 then IntPow := T else IntPow := 1.0 / T;
  end;

  function Pow(X, Y : Float) : Float;
  { Computes X^Y = Exp(Y * Ln(X)), for X >= 0 }
  var
    YLnX : Float;
  begin
    MathErr := FN_OK;
    if X < 0.0 then
      begin
        Pow := DefaultVal(DOMAIN);
        Exit;
      end;

    if X = 0.0 then
      begin
        if Y = 0.0 then       { 0^0 = lim  x^x = 1 }
          Pow := 1.0          {       x->0         }
        else if Y > 0.0 then
          Pow := 0.0          { 0^Y = 0 }
        else
          Pow := DefaultVal(SING);
        Exit;
      end;

    if Y = 0.0 then
      begin
        Pow := 1.0;
        Exit;
      end;

  {$IFDEF CPU387}
    YLnX :=  Y * fln(X);
  {$ELSE}
    YLnX := Y * System.Ln(X);
  {$ENDIF}

    if YLnX < MINLOG then
      Pow := DefaultVal(UNDERFLOW)
    else if YLnX > MAXLOG then
      Pow := DefaultVal(OVERFLOW)
    else
    {$IFDEF CPU387}
      Pow := fexp(YLnX);
    {$ELSE}
      Pow := System.Exp(YLnX);
    {$ENDIF}
  end;

  function Pythag(X, Y : Float) : Float;
  { Computes Sqrt(X^2 + Y^2) without destructive underflow or overflow }
  var
    AbsX, AbsY : Float;
  begin
    MathErr := FN_OK;
    AbsX := Abs(X);
    AbsY := Abs(Y);
    if AbsX > AbsY then
      Pythag := AbsX * Sqrt(1.0 + Sqr(AbsY / AbsX))
    else if AbsY = 0.0 then
      Pythag := 0.0
    else
      Pythag := AbsY * Sqrt(1.0 + Sqr(AbsX / AbsY));
  end;

  function FixAngle(Theta : Float) : Float;
  begin
    MathErr := FN_OK;
    while Theta > PI do
      Theta := Theta - TWOPI;
    while Theta <= - PI do
      Theta := Theta + TWOPI;
    FixAngle := Theta;
  end;

  function Sin(X : Float) : Float;
  begin
    MathErr := FN_OK;
  {$IFDEF CPU387}
    Sin := fsin(X);
  {$ELSE}
    Sin := System.Sin(X);
  {$ENDIF}
  end;

  function Cos(X : Float) : Float;
  begin
    MathErr := FN_OK;
  {$IFDEF CPU387}
    Cos := fcos(X);
  {$ELSE}
    Cos := System.Cos(X);
  {$ENDIF}
  end;

  procedure SinCos(X : Float; var SinX, CosX : Float);
  begin
    MathErr := FN_OK;
  {$IFDEF CPU387}
    fsincos(X, SinX, CosX);
  {$ELSE}
    SinX := System.Sin(X);
    CosX := System.Cos(X);
  {$ENDIF}
  end;

  function Tan(X : Float) : Float;
  var
    SinX, CosX : Float;
  begin
    MathErr := FN_OK;
    SinCos(X, SinX, CosX);
    if CosX = 0.0 then
      Tan := Sgn(SinX) * DefaultVal(SING)
    else
      Tan := SinX / CosX;
  end;

  function ArcSin(X : Float) : Float;
  begin
    MathErr := FN_OK;
    if (X < - 1.0) or (X > 1.0) then
      ArcSin := DefaultVal(DOMAIN)
    else if X = 1.0 then
      ArcSin := PIDIV2
    else if X = - 1.0 then
      ArcSin := - PIDIV2
    else
    {$IFDEF CPU387}
      ArcSin := farcsin(X);
    {$ELSE}
      ArcSin := System.ArcTan(X / Sqrt(1.0 - Sqr(X)));
    {$ENDIF}
  end;

  function ArcCos(X : Float) : Float;
  begin
    MathErr := FN_OK;
    if (X < - 1.0) or (X > 1.0) then
      ArcCos := DefaultVal(DOMAIN)
    else if X = 1.0 then
      ArcCos := 0.0
    else if X = - 1.0 then
      ArcCos := PI
    else
    {$IFDEF CPU387}
      ArcCos := farccos(X);
    {$ELSE}
      ArcCos := PIDIV2 - System.ArcTan(X / Sqrt(1.0 - Sqr(X)));
    {$ENDIF}
  end;

  function ArcTan(X : Float) : Float;
  begin
    MathErr := FN_OK;
  {$IFDEF CPU387}
    ArcTan := farctan(X);
  {$ELSE}
    ArcTan := System.ArcTan(X);
  {$ENDIF}
  end;

  function ArcTan2(X, Y : Float) : Float;
  {$IFDEF CPU387}
  begin
    MathErr := FN_OK;
    ArcTan2 := farctan2(X, Y);
  {$ELSE}
  var
    Theta : Float;
  begin
    MathErr := FN_OK;
    if X = 0.0 then
      if Y = 0.0 then
        ArcTan2 := 0.0
      else if Y > 0.0 then
        ArcTan2 := PIDIV2
      else
        ArcTan2 := - PIDIV2
    else
      begin
        Theta := ArcTan(Y / X);   { 4th/1st quadrant -PI/2..PI/2 }
        if X < 0.0 then           { 2nd/3rd quadrants }
          if Y >= 0.0 then
            Theta := Theta + PI   { 2nd quadrant:  PI/2..PI }
          else
            Theta := Theta - PI;  { 3rd quadrant: -PI..-PI/2 }
        ArcTan2 := Theta;
      end;
  {$ENDIF}
  end;

  function Sinh(X : Float) : Float;
  var
    ExpX : Float;
  begin
    MathErr := FN_OK;
    if (X < MINLOG) or (X > MAXLOG) then
      Sinh := Sgn(X) * DefaultVal(OVERFLOW)
    else
      begin
      {$IFDEF CPU387}
        ExpX := fexp(X);
      {$ELSE}
        ExpX := System.Exp(X);
      {$ENDIF}
        Sinh := 0.5 * (ExpX - 1.0 / ExpX);
      end;
  end;

  function Cosh(X : Float) : Float;
  var
    ExpX : Float;
  begin
    MathErr := FN_OK;
    if (X < MINLOG) or (X > MAXLOG) then
      Cosh := DefaultVal(OVERFLOW)
    else
      begin
      {$IFDEF CPU387}
        ExpX := fexp(X);
      {$ELSE}
        ExpX := System.Exp(X);
      {$ENDIF}
        Cosh := 0.5 * (ExpX + 1.0 / ExpX);
      end;
  end;

  procedure SinhCosh(X : Float; var SinhX, CoshX : Float);
  var
    ExpX, ExpMinusX : Float;
  begin
    MathErr := FN_OK;
    if (X < MINLOG) or (X > MAXLOG) then
      begin
        CoshX := DefaultVal(OVERFLOW);
        SinhX := Sgn(X) * CoshX;
      end
    else
      begin
      {$IFDEF CPU387}
        ExpX := fexp(X);
      {$ELSE}
        ExpX := System.Exp(X);
      {$ENDIF}
        ExpMinusX := 1.0 / ExpX;
        SinhX := 0.5 * (ExpX - ExpMinusX);
        CoshX := 0.5 * (ExpX + ExpMinusX);
      end;
  end;

  function Tanh(X : Float) : Float;
  var
    SinhX, CoshX : Float;
  begin
    SinhCosh(X, SinhX, CoshX);
    Tanh := SinhX / CoshX;
  end;

  function ArcSinh(X : Float) : Float;
  begin
    MathErr := FN_OK;
  {$IFDEF CPU387}
    ArcSinH := fln(X + Sqrt(Sqr(X) + 1.0));
  {$ELSE}
    ArcSinh := System.Ln(X + Sqrt(Sqr(X) + 1.0));
  {$ENDIF}
  end;

  function ArcCosh(X : Float) : Float;
  begin
    MathErr := FN_OK;
    if X < 1.0 then
      ArcCosh := DefaultVal(DOMAIN)
    else
    {$IFDEF CPU387}
      ArcCosH := fln(X + Sqrt(Sqr(X) - 1.0));
    {$ELSE}
      ArcCosh := System.Ln(X + Sqrt(Sqr(X) - 1.0));
    {$ENDIF}
  end;

  function ArcTanh(X : Float) : Float;
  begin
    MathErr := FN_OK;
    if (X < - 1.0) or (X > 1.0) then
      ArcTanh := DefaultVal(DOMAIN)
    else if (X = - 1.0) or (X = 1.0) then
      ArcTanh := Sgn(X) * DefaultVal(SING)
    else
    {$IFDEF CPU387}
      ArcTanH := 0.5 * fln((1.0 + X) / (1.0 - X));
    {$ELSE}
      ArcTanh := 0.5 * System.Ln((1.0 + X) / (1.0 - X));
    {$ENDIF}
  end;

  function SgnGamma(X : Float) : Integer;
  begin
    if X > 0.0 then
      SgnGamma := 1
    else if Odd(Trunc(Abs(X))) then
      SgnGamma := 1
    else
      SgnGamma := - 1;
  end;

  function Stirf(X : Float) : Float;
  { Stirling's formula for the gamma function
    Gamma(x) = Sqrt(2*Pi) x^(x-.5) exp(-x) (1 + 1/x P(1/x)) }
  var
    W, P : Float;
  begin
    W := 1.0 / X;
    if X > 1024.0 then
      begin
        P := 6.97281375836585777429E-5 * W + 7.84039221720066627474E-4;
        P := P * W - 2.29472093621399176955E-4;
        P := P * W - 2.68132716049382716049E-3;
        P := P * W + 3.47222222222222222222E-3;
        P := P * W + 8.33333333333333333333E-2;
      end
    else
      begin
        P := 7.147391378143610789273E-4 * W - 2.363848809501759061727E-5;
        P := P * W - 5.950237554056330156018E-4;
        P := P * W + 6.989332260623193171870E-5;
        P := P * W + 7.840334842744753003862E-4;
        P := P * W - 2.294719747873185405699E-4;
        P := P * W - 2.681327161876304418288E-3;
        P := P * W + 3.472222222230075327854E-3;
        P := P * W + 8.333333333333331800504E-2;
      end;
    Stirf := SQRT2PI * Exp((X - 0.5) * Ln(X) - X) * (1.0 + W * P);
  end;

  function GamSmall(X1, Z : Float) : Float;
  { Gamma function for small values of the argument }
  var
    P : Float;
  begin
    if X1 = 0.0 then
      GamSmall := DefaultVal(SING)
    else if X1 < 0.0 then
      begin
        X1 := - X1;
        P := 1.133374167243894382010E-3 * X1 + 7.220837261893170325704E-3;
        P := P * X1 + 9.621911155035976733706E-3;
        P := P * X1 - 4.219773343731191721664E-2;
        P := P * X1 - 1.665386113944413519335E-1;
        P := P * X1 - 4.200263503402112910504E-2;
        P := P * X1 + 6.558780715202536547116E-1;
        P := P * X1 + 5.772156649015328608727E-1;
        P := P * X1 - 1.0;
      end
    else
      begin
        P := - 1.193945051381510095614E-3 * X1 + 7.220599478036909672331E-3;
        P := P * X1 - 9.622023360406271645744E-3;
        P := P * X1 - 4.219773360705915470089E-2;
        P := P * X1 + 1.665386113720805206758E-1;
        P := P * X1 - 4.200263503403344054473E-2;
        P := P * X1 - 6.558780715202540684668E-1;
        P := P * X1 + 5.772156649015328608253E-1;
        P := P * X1 + 1.0;
      end;
    GamSmall := Z / (X1 * P);
  end;

  function StirfL(X : Float) : Float;
  { Approximate Ln(Gamma) by Stirling's formula, for X >= 13 }
  var
    P, Q, W : Float;
  begin
    Q := Ln(X) * (X - 0.5) - X;
    Q := Q + LNSQRT2PI;
    if X > 1.0E+10 then
      StirfL := Q
    else
      begin
        W := 1.0 / Sqr(X);
        P := 4.885026142432270781165E-3 * W - 1.880801938119376907179E-3;
        P := P * W + 8.412723297322498080632E-4;
        P := P * W - 5.952345851765688514613E-4;
        P := P * W + 7.936507795855070755671E-4;
        P := P * W - 2.777777777750349603440E-3;
        P := P * W + 8.333333333333331447505E-2;
        StirfL := Q + P / X;
      end;
  end;

  function Gamma(X : Float) : Float;
  var
    SgnGam, N : Integer;
    A, P, Q, X1, Z : Float;
  begin
    MathErr := FN_OK;
    SgnGam := SgnGamma(X);

    if (X = 0.0) or ((X < 0.0) and (Frac(X) = 0.0)) then
      begin
        Gamma := SgnGam * DefaultVal(SING);
        Exit;
      end;

    if X > MAXGAM then
      begin
        Gamma := DefaultVal(OVERFLOW);
        Exit;
      end;

    A := Abs(X);
    if A > 13.0 then
      begin
        if X < 0.0 then
          begin
            N := Trunc(A);
            Z := A - N;
            if Z > 0.5 then
              begin
                N := N + 1;
                Z := A - N;
              end;
            Z := Abs(A * Sin(PI * Z)) * Stirf(A);
            if Z <= PI / MAXNUM then
              begin
                Gamma := SgnGam * DefaultVal(OVERFLOW);
                Exit;
              end;
            Z := PI / Z;
          end
        else
          Z := Stirf(X);
        Gamma := SgnGam * Z;
      end
    else
      begin
        Z := 1.0;
        X1 := X;
        while X1 >= 3.0 do
          begin
            X1 := X1 - 1.0;
            Z := Z * X1;
          end;
        while X1 < - 0.03125 do
          begin
            Z := Z / X1;
            X1 := X1 + 1.0;
          end;
        if X1 <= 0.03125 then
          Gamma := GamSmall(X1, Z)
        else
          begin
            while X1 < 2.0 do
              begin
                Z := Z / X1;
                X1 := X1 + 1.0;
              end;
            if (X1 = 2.0) or (X1 = 3.0) then
              Gamma := Z
            else
              begin
                X1 := X1 - 2.0;
                P := 4.212760487471622013093E-5 * X1 + 4.542931960608009155600E-4;
                P := P * X1 + 4.092666828394035500949E-3;
                P := P * X1 + 2.385363243461108252554E-2;
                P := P * X1 + 1.113062816019361559013E-1;
                P := P * X1 + 3.629515436640239168939E-1;
                P := P * X1 + 8.378004301573126728826E-1;
                P := P * X1 + 1.000000000000000000009E0;
                Q := - 1.397148517476170440917E-5 * X1 + 2.346584059160635244282E-4;
                Q := Q * X1 - 1.237799246653152231188E-3;
                Q := Q * X1 - 7.955933682494738320586E-4;
                Q := Q * X1 + 2.773706565840072979165E-2;
                Q := Q * X1 - 4.633887671244534213831E-2;
                Q := Q * X1 - 2.243510905670329164562E-1;
                Q := Q * X1 + 4.150160950588455434583E-1;
                Q := Q * X1 + 9.999999999999999999908E-1;
                Gamma := Z * P / Q;
              end;
          end;
      end;
  end;

  function LnGamma(X : Float) : Float;
  var
    N : Integer;
    A, P, Q, X1, Z : Float;
  begin
    MathErr := FN_OK;

    if (X = 0.0) or ((X < 0.0) and (Frac(X) = 0.0)) then
      begin
        LnGamma := DefaultVal(SING);
        Exit;
      end;

    if X > MAXLGM then
      begin
        LnGamma := DefaultVal(OVERFLOW);
        Exit;
      end;

    A := Abs(X);
    if A > 34.0 then
      begin
        if X < 0.0 then
          begin
            N := Trunc(A);
            Z := A - N;
            if Z > 0.5 then
              begin
                N := N + 1;
                Z := N - A;
              end;
            Z := A * Sin(PI * Z);
            if Z = 0.0 then
              begin
                LnGamma := DefaultVal(OVERFLOW);
                Exit;
              end;
            Z := LNPI - Ln(Z) - StirfL(A);
          end
        else
          Z := StirfL(X);
        LnGamma := Z;
      end
    else if X < 13.0 then
      begin
        Z := 1.0;
        X1 := X;
        while X1 >= 3 do
          begin
            X1 := X1 - 1.0;
            Z := Z * X1;
          end;
        while X1 < 2.0 do
          begin
            if Abs(X1) <= 0.03125 then
              begin
                LnGamma := Ln(Abs(GamSmall(X1, Z)));
                Exit;
              end;
            Z := Z / X1;
            X1 := X1 + 1.0;
          end;
        if Z < 0.0 then Z := - Z;
        if X1 = 2.0 then
          LnGamma := Ln(Z)
        else
          begin
            X1 := X1 - 2.0;
            P := - 2.163690827643812857640E3 * X1 - 8.723871522843511459790E4;
            P := P * X1 - 1.104326814691464261197E6;
            P := P * X1 - 6.111225012005214299996E6;
            P := P * X1 - 1.625568062543700591014E7;
            P := P * X1 - 2.003937418103815175475E7;
            P := P * X1 - 8.875666783650703802159E6;
            Q := X1 - 5.139481484435370143617E2;
            Q := Q * X1 - 3.403570840534304670537E4;
            Q := Q * X1 - 6.227441164066219501697E5;
            Q := Q * X1 - 4.814940379411882186630E6;
            Q := Q * X1 - 1.785433287045078156959E7;
            Q := Q * X1 - 3.138646407656182662088E7;
            Q := Q * X1 - 2.099336717757895876142E7;
            LnGamma := Ln(Z) + X1 * P / Q;
          end;
      end
    else
      LnGamma := StirfL(X);
  end;

  function Fact(N : Integer) : Float;
  begin
    MathErr := FN_OK;
    if N < 0 then
      Fact := DefaultVal(DOMAIN)
    else if N > MAXFAC then
      Fact := DefaultVal(OVERFLOW)
    else if N <= NFACT then
      Fact := FactArray[N]
    else
      Fact := Gamma(N + 1);
  end;

  function Binomial(N, K : Integer) : Float;
  var
    I, N1 : Integer;
    Prod : Float;
  begin
    MathErr := FN_OK;
    if K < 0 then
      Binomial := 0.0
    else if (K = 0) or (K = N) then
      Binomial := 1.0
    else if (K = 1) or (K = N - 1) then
      Binomial := N
    else
      begin
        if K > N - K then K := N - K;
        N1 := Succ(N);
        Prod := N;
        for I := 2 to K do
          Prod := Prod * ((N1 - I) / I);
        Binomial := Int(0.5 + Prod);
      end;
  end;

  function Beta(X, Y : Float) : Float;
  { Computes Beta(X, Y) = Gamma(X) * Gamma(Y) / Gamma(X + Y) }
  var
    Lx, Ly, Lxy : Float;
    SgnBeta : Integer;
  begin
    MathErr := FN_OK;
    SgnBeta := SgnGamma(X) * SgnGamma(Y) * SgnGamma(X + Y);
    Lxy := LnGamma(X + Y);
    if MathErr <> FN_OK then
      begin
        Beta := 0.0;
        Exit;
      end;
    Lx := LnGamma(X);
    if MathErr <> FN_OK then
      begin
        Beta := SgnBeta * MAXNUM;
        Exit;
      end;
    Ly := LnGamma(Y);
    if MathErr <> FN_OK then
      begin
        Beta := SgnBeta * MAXNUM;
        Exit;
      end;
    Beta := SgnBeta * Exp(Lx + Ly - Lxy);
  end;

  function GammaFrac(A, X : Float) : Float;
  { Incomplete Gamma function : continued fraction expansion }
  var
    I : Integer;
    An, B, C, D, Del, H : Float;
  begin
    B := X + 1.0 - A;
    C := 1.0 / MINNUM;
    D := 1.0 / B;
    H := D;
    I := 0;
    repeat
      Inc(I);
      An := - I * (I - A);
      B := B + 2.0;
      D := An * D + B;
      if Abs(D) < MINNUM then D := MINNUM;
      C := B + An / C;
      if Abs(C) < MINNUM then C := MINNUM;
      D := 1.0 / D;
      Del := D * C;
      H := H * Del;
    until Abs(Del - 1.0) < MACHEP;
    GammaFrac := Exp(- X + A * Ln(X) - LnGamma(A)) * H;
  end;

  function GammaSer(A, X : Float) : Float;
  { Incomplete Gamma function : series expansion }
  var
    Ap, Del, Sum : Float;
  begin
    Ap := A;
    Sum := 1.0 / A;
    Del := Sum;
    repeat
      Ap := Ap + 1.0;
      Del := Del * X / Ap;
      Sum := Sum + Del;
    until Abs(Del) < Abs(Sum) * MACHEP;
    GammaSer := Sum * Exp(- X + A * Ln(X) - LnGamma(A));
  end;

  function IGamma(A, X : Float) : Float;
  begin
    MathErr := FN_OK;
    if (A <= 0.0) or (X < 0.0) then
      IGamma := DefaultVal(DOMAIN)
    else if X = 0.0 then
      IGamma := 0.0
    else if X < A + 1.0 then
      IGamma := GammaSer(A, X)
    else
      IGamma := 1.0 - GammaFrac(A, X);
  end;

  function JGamma(A, X : Float) : Float;
  { Complement of IGamma }
  begin
    MathErr := FN_OK;
    if (A <= 0.0) or (X < 0.0) then
      JGamma := DefaultVal(DOMAIN)
    else if X = 0.0 then
      JGamma := 1.0
    else if X < A + 1.0 then
      JGamma := 1.0 - GammaSer(A, X)
    else
      JGamma := GammaFrac(A, X);
  end;

  function BetaFrac(A, B, X : Float) : Float;
  { Incomplete Beta function : continued fraction expansion }
  var
    M, M2 : Integer;
    AA, C, D, Del, H : Float;
    Qab, Qam, Qap : Float;
  begin
    Qab := A + B;
    Qap := A + 1.0;
    Qam := A - 1.0;
    C := 1.0;
    D := 1.0 - Qab * X / Qap;
    if Abs(D) < MINNUM then D := MINNUM;
    D := 1.0 / D;
    H := D;
    M := 0;
    repeat
      Inc(M);
      M2 := 2 * M;
      AA := M * (B - M) * X / ((Qam + M2) * (A + M2));
      D := 1.0 + AA * D;
      if Abs(D) < MINNUM then D := MINNUM;
      C := 1.0 + AA / C;
      if Abs(C) < MINNUM then C := MINNUM;
      D := 1.0 / D;
      H := H * D * C;
      AA := - (A + M) * (Qab + M) * X / ((A + M2) * (Qap + M2));
      D := 1.0 + AA * D;
      if Abs(D) < MINNUM then D := MINNUM;
      C := 1.0 + AA / C;
      if Abs(C) < MINNUM then C := MINNUM;
      D := 1.0 / D;
      Del := D * C;
      H := H * Del;
    until Abs(Del - 1.0) < MACHEP;
    BetaFrac := H;
  end;

  function IBeta(A, B, X : Float) : Float;
  var
    Bt : Float;
  begin
    MathErr := FN_OK;
    if (A <= 0.0) or (B <= 0.0) or (X < 0.0) or (X > 1.0) then
      IBeta := DefaultVal(DOMAIN)
    else if (X = 0.0) or (X = 1.0) then
      IBeta := X
    else
      begin
        Bt := Exp(LnGamma(A + B) - LnGamma(A) - LnGamma(B) +
                  A * Ln(X) + B * Ln(1.0 - X));
        if X < (A + 1.0) / (A + B + 2.0) then
          IBeta := Bt * BetaFrac(A, B, X) / A
        else
          IBeta := 1.0 - Bt * BetaFrac(B, A, 1.0 - X) / B;
      end;
  end;

  function Erf(X : Float) : Float;
  begin
    if X < 0.0 then
      Erf := - IGamma(0.5, Sqr(X))
    else
      Erf := IGamma(0.5, Sqr(X));
  end;

  function Erfc(X : Float) : Float;
  begin
    if X < 0.0 then
      Erfc := 1.0 + IGamma(0.5, Sqr(X))
    else
      Erfc := JGamma(0.5, Sqr(X));
  end;

  function PNorm(X : Float) : Float;
  var
    A : Float;
  begin
    A := Abs(X);
    MathErr := FN_OK;
    if A = 0.0 then
      PNorm := 1.0
    else if A < 1.0 then
      PNorm := 1.0 - Erf(A * SQRT2DIV2)
    else
      PNorm := Erfc(A * SQRT2DIV2);
  end;

  function PStudent(Nu : Integer; X : Float) : Float;
  begin
    MathErr := FN_OK;
    if Nu < 1 then
      PStudent := DefaultVal(DOMAIN)
    else
      PStudent := IBeta(0.5 * Nu, 0.5, Nu / (Nu + Sqr(X)));
  end;

  function PKhi2(Nu : Integer; X : Float) : Float;
  begin
    MathErr := FN_OK;
    if Nu < 1 then
      PKhi2 := DefaultVal(DOMAIN)
    else
      PKhi2 := JGamma(0.5 * Nu, 0.5 * X);
  end;

  function PSnedecor(Nu1, Nu2 : Integer; X : Float) : Float;
  begin
    MathErr := FN_OK;
    if (Nu1 < 1) or (Nu2 < 1) then
      PSnedecor := DefaultVal(DOMAIN)
    else
      PSnedecor := IBeta(0.5 * Nu2, 0.5 * Nu1, Nu2 / (Nu2 + Nu1 * X));
  end;

  procedure RMarIn(Seed1, Seed2 : Integer);
  var
    I, J, K, L, M, II, JJ : Integer;
    SS, TT : LongInt;
  begin
    { Compute I, J, K and L from Seed1 and Seed2
      2 <= I < 179, 2 <= J < 179, 1 <= K < 179, 0 <= L < 169 }
    I := ((Seed1 div 177) mod 177) + 2;
    J := (Seed1 mod 177) + 2;
    K := ((Seed2 div 169) mod 178) + 1;
    L := Seed2 mod 169;

    { II cycles through all elements of X, the random look-up table }
    for II := LASTDIM downto 1 do
      begin
        { Initialize SS (the bit accumulator) }
        { Initialize TT (the single-bit mask) }
        { JJ cycles through 24 bits }
        SS := 0; TT := NO_NUMBERS; JJ := 1;
        for JJ := 1 to 24 do
          begin
            { 0 <= M < 179 }
            M := ((I * J mod 179) * K) mod 179;

            { Shift J, K and M into I, J and K }
            I := J; J := K; K := M;

            { 0 <= L < 169 }
            L := (53 * L + 1) mod 169;

            { TT shifts to next bit }
            TT := TT shr 1;

            { Accumulate TT into SS, about half the time }
            if ((L * M) and 32) <> 0 then SS := SS or TT;
          end;

        { Put SS into look-up table }
        X[II] := SS;
      end;

    { Assign remaining globals: J_ptr, K_ptr and C }
    J_ptr := 1;
    K_ptr := FIRSTDIM;
    C := 362436;
  end;

  function IRanMar : LongInt;
  var
    Uni : LongInt;
  begin
    { Uni is difference between two table values }
    Uni := X[J_ptr] - X[K_ptr];

    { Make sure 0 <= Uni < NO_NUMBERS }
    if Uni < 0 then Uni := Uni + NO_NUMBERS;

    { Assign Uni into look-up table }
    X[J_ptr] := Uni;

    { Point J_ptr and K_ptr at new look-up table values }
    J_ptr := J_ptr + 1; if J_ptr > LASTDIM then J_ptr := 1;
    K_ptr := K_ptr + 1; if K_ptr > LASTDIM then K_ptr := 1;

    { Modify C; 0 <= C <NO_NUMBERS-CD or CM-CD <= C < CM }
    { There are three overlap positions between CM-CD and NO_NUMBERS-CD, }
    { and three unavailable positions between CM and NO_NUMBERS }
    C := C - CD; if C < 0 then C := C + CM;

    { Modify Uni, make sure 0 <= uni < NO_NUMBERS }
    Uni := Uni - C; if Uni < 0 then Uni := Uni + NO_NUMBERS;
    IRanMar := Uni;
  end;

  function RanMar : Float;
  var
    X : LongInt;
  begin
    X := IRanMar;
    if X = 0 then X := 1;
    RanMar := X / NO_NUMBERS;
  end;

  function RanGauss(Mu, Sigma : Float) : Float;
  { This routine computes 2 gaussian random numbers,
    returns one and saves the other for the next call. }
  const
    Gauss_Save : Float = 0.0;     { Saves a random number }
    Gauss_Set : Boolean = False;  { Flags if a number has been saved }
  var
    Fac, R, U, V1, V2 : Float;
  begin
    if not Gauss_Set then
      begin
        repeat
          V1 := 2.0 * RanMar - 1.0;
          V2 := 2.0 * RanMar - 1.0;
          R := Sqr(V1) + Sqr(V2);
        until R < 1.0;
        Fac := Sqrt(- 2.0 * Ln(R) / R);
        U := V1 * Fac;
        Gauss_Save := V2 * Fac;
      end
    else
      U := Gauss_Save;
    RanGauss := Mu + U * Sigma;
    Gauss_Set := not Gauss_Set;
  end;

begin
  { Initialize MathErr }
  MathErr := FN_OK;

  { Initialize random number generator }
  RMarIn(1802, 9373);

  { Compute factorials }
  FactArray[0] := 1.0;
  FactArray[1] := 1.0;
  FactArray[2] := 2.0;
  for I := 3 to NFACT do
    FactArray[I] := FactArray[I - 1] * I;
end.
