{ **********************************************************************
  *                            Unit OPTIM.PAS                          *
  *                              Version 1.4                           *
  *                      (c) J. Debord, August 1997                    *
  **********************************************************************
  This unit implements the Marquardt and Simplex methods for function
  minimization.

  The address of the objective function to be minimized is stored in
  the global variable ObjFuncAddr. You must pass the address of your
  function to this variable before calling any of the routines defined
  here. See the sample programs HESSIAN.PAS and MINFUNC.PAS for details.
  **********************************************************************
  References:
  1) D. W. MARQUARDT, J. Soc. Indust. Appl. Math., 1963, 11, 431-441
  2) J. A. NELDER & R. MEAD, Comput. J., 1964, 7, 308-313
  3) R. O'NEILL, Appl.Statist., 1971, 20, 338-345
  ********************************************************************** }

unit Optim;

interface

uses
  FMath, Matrices;

{ **************************** Error codes ***************************** }

const
  BIG_LAMBDA = - 2;  { Too high Marquardt's parameter }
  NON_CONV   = - 3;  { Non-convergence }

{ ********************** Addresses of procedures *********************** }

var
  { Address of the objective function to be minimized }
  ObjFuncAddr : Pointer;

  { Address of the procedure which computes the Gradient vector
    and Hessian matrix of the objective function }
  HessProcAddr : Pointer;

{ *********************** Minimization routines ************************ }

procedure HessGrad(X : PVector; Lbound, Ubound : Integer;
                   G : PVector; H : PMatrix);
{ ----------------------------------------------------------------------
  Computes Gradient vector & Hessian matrix by numerical differentiation
  ----------------------------------------------------------------------
  Input parameters  : X       = vector of independent variables
                      Lbound,
                      Ubound  = indices of first and last values of X
  ----------------------------------------------------------------------
  Output parameters : G       = gradient vector
                      H       = hessian matrix
  ---------------------------------------------------------------------- }

function Marquardt(X : PVector; Lbound, Ubound, MaxIter : Integer;
                   Tol : Float; var F_min : Float;
                   H_inv : PMatrix) : Integer;
{ ----------------------------------------------------------------------
  Function minimization by Marquardt's method
  ----------------------------------------------------------------------
  Input parameters  : X, Lbound, Ubound = as in HessGrad
                      MaxIter           = maximum number of iterations
                      Tol               = required precision
  ----------------------------------------------------------------------
  Output parameters : X     = refined minimum coordinates
                      F_min = function value at minimum
                      H_inv = inverse hessian matrix
  ----------------------------------------------------------------------
  Possible results  : MAT_OK
                      MAT_SINGUL
                      BIG_LAMBDA
                      NON_CONV
  ---------------------------------------------------------------------- }

function Simplex(X : PVector; Lbound, Ubound, MaxIter : Integer;
                 Tol : Float; var F_min : Float) : Integer;
{ ----------------------------------------------------------------------
  Function minimization by the simplex method of Nelder and Mead
  ----------------------------------------------------------------------
  Parameters : as in Marquardt
  ----------------------------------------------------------------------
  Possible results : MAT_OK
                     NON_CONV
  ---------------------------------------------------------------------- }


implementation

const
{$IFOPT N+}
  EPS = 1.0E-6;  { Fractional increment for numerical derivation }
{$ELSE}
  EPS = 2.0E-4;
{$ENDIF}

  function ObjFunc(X : PVector) : Float;
  { Formal function called by the minimization routines }
  inline($FF / $1E / ObjFuncAddr);

  procedure HessProc(X : PVector; Lbound, Ubound : Integer;
                     G : PVector; H : PMatrix);
  { Formal procedure called by the Marquardt routine }
  inline($FF / $1E / HessProcAddr);

  {$F+}
  procedure HessGrad(X : PVector; Lbound, Ubound : Integer;
                     G : PVector; H : PMatrix);
  var
    I, J : Integer;
    Temp1, Temp2, F, F2plus : Float;
    Delta, Xminus, Xplus, Fminus, Fplus : PVector;
  begin
    DimVector(Delta, Ubound);   { Increments   }
    DimVector(Xminus, Ubound);  { X - Delta    }
    DimVector(Xplus, Ubound);   { X + Delta    }
    DimVector(Fminus, Ubound);  { F(X - Delta) }
    DimVector(Fplus, Ubound);   { F(X + Delta) }

    F := ObjFunc(X);

    for I := Lbound to Ubound do
      begin
        Delta^[I] := EPS * Abs(X^[I]);
        Xplus^[I] := X^[I] + Delta^[I];
        Xminus^[I] := X^[I] - Delta^[I];
      end;

    for I := Lbound to Ubound do
      begin
        Temp1 := X^[I];
        X^[I] := Xminus^[I];
        Fminus^[I] := ObjFunc(X);
        X^[I] := Xplus^[I];
        Fplus^[I] := ObjFunc(X);
        X^[I] := Temp1;
      end;

    for I := Lbound to Ubound do
      begin
        G^[I] := (Fplus^[I] - Fminus^[I]) / (2.0 * Delta^[I]);
        H^[I]^[I] := (Fplus^[I] + Fminus^[I] - 2.0 * F) / Sqr(Delta^[I]);
      end;

    for I := Lbound to Pred(Ubound) do
      begin
        Temp1 := X^[I];
        X^[I] := Xplus^[I];
        for J := Succ(I) to Ubound do
          begin
            Temp2 := X^[J];
            X^[J] := Xplus^[J];
            F2plus := ObjFunc(X);
            H^[I]^[J] := (F2plus - Fplus^[I] - Fplus^[J] + F) /
                                             (Delta^[I] * Delta^[J]);
            H^[J]^[I] := H^[I]^[J];
            X^[J] := Temp2;
          end;
        X^[I] := Temp1;
      end;

    DelVector(Delta, Ubound);
    DelVector(Xminus, Ubound);
    DelVector(Xplus, Ubound);
    DelVector(Fminus, Ubound);
    DelVector(Fplus, Ubound);
  end;
  {$F-}

  function Marquardt(X : PVector; Lbound, Ubound, MaxIter : Integer;
                     Tol : Float; var F_min : Float;
                     H_inv : PMatrix) : Integer;
  const
    LAMBDAMIN = 1.0E-2;  { Lowest lambda value }
    LAMBDAMAX = 1.0E+3;  { Highest lambda value }
    TOLMAX    = 1.0E+1;  { Highest acceptable correction }
  var
    Lambda,
    Lambda1 : Float;     { Marquardt's lambda }
    Corr,
    AbsCorr,
    MaxCorr : Float;     { Correcting terms }
    I       : Integer;   { Loop variable }
    Iter    : Integer;   { Iteration number }
    ErrCode : Integer;   { Error code }
    Ok      : Boolean;   { Flags a successful iteration }
    G       : PVector;   { Gradient vector }
    H       : PMatrix;   { Hessian matrix }
    A       : PMatrix;   { Modified Hessian matrix }
    Delta   : PVector;   { Correcting terms }
    X1      : PVector;   { New coordinates }
    F1      : Float;     { New minimum }
  begin
    Lambda := LAMBDAMIN;
    MaxCorr := TOLMAX;
    ErrCode := MAT_OK;

    DimVector(G, Ubound);
    DimMatrix(H, Ubound, Ubound);
    DimMatrix(A, Ubound, Ubound);
    DimVector(Delta, Ubound);
    DimVector(X1, Ubound);

    Iter := 1;
    F_min := ObjFunc(X);

    repeat
      { Compute Gradient and Hessian }
      HessProc(X, Lbound, Ubound, G, H);
      CopyMatrix(A, H, Lbound, Lbound, Ubound, Ubound);

      { If minimum reached, do one more iteration with lambda set to 0
        i.e. perform a pure Newton-Raphson iteration }
      if MaxCorr < Tol then Lambda := 0.0;

      repeat
        { Multiply each diagonal term of H by (1 + lambda) }
        if Lambda > 0.0 then
          begin
            Lambda1 := 1.0 + Lambda;
            for I := Lbound to Ubound do
              A^[I]^[I] := Lambda1 * H^[I]^[I];
          end;

        { Solve system A.Delta = G }
        ErrCode := GaussJordan(A, G, Lbound, Ubound, H_inv, Delta);

        if ErrCode = MAT_OK then
          begin
            { Compute correcting terms by Newton-Raphson formula
              Check that corrections do not exceed TOLMAX
              Compute highest correction to test convergence }
            MaxCorr := 0.0;
            for I := Lbound to Ubound do
              begin
                Corr := Delta^[I] / X^[I];
                AbsCorr := Abs(Corr);
                if AbsCorr > TOLMAX then Corr := Sgn(Corr) * TOLMAX;
                if AbsCorr > MaxCorr then MaxCorr := AbsCorr;
                X1^[I] := X^[I] * (1.0 - Corr);
              end;

            { Compute new function value. If function did not decrease,
              increase Lambda without exceeding LAMBDAMAX }
            F1 := ObjFunc(X1);
            Ok := (F1 < (1.0 + Tol) * F_min);
            if not Ok then Lambda := 10.0 * Lambda;
            if Lambda > LAMBDAMAX then ErrCode := BIG_LAMBDA;
          end;
      until Ok or (Lambda = 0.0) or (ErrCode <> MAT_OK);

      { Update variables for next iteration }
      CopyVector(X, X1, Lbound, Ubound);
      Lambda := 0.1 * Lambda;
      F_min := F1;
      Inc(Iter);
      if Iter > MaxIter then ErrCode := NON_CONV;
    until (Lambda = 0.0) or (ErrCode <> MAT_OK);

    Marquardt := ErrCode;

    DelVector(G, Ubound);
    DelMatrix(H, Ubound, Ubound);
    DelMatrix(A, Ubound, Ubound);
    DelVector(Delta, Ubound);
    DelVector(X1, Ubound);
  end;

  function Simplex(X : PVector; Lbound, Ubound, MaxIter : Integer;
                   Tol : Float; var F_min : Float) : Integer;
  const
    { Multiplicator used to construct the initial simplex }
    Delta = 1.50;
  var
    P       : PMatrix;  { Simplex coordinates }
    F       : PVector;  { Function values }
    Pbar    : PVector;  { Centroid coordinates }
    Pstar,
    P2star  : PVector;  { New vertices }
    Ystar,
    Y2star  : Float;    { New function values }
    F0      : Float;    { Function value at minimum }
    N       : Integer;  { Number of parameters }
    M       : Integer;  { Index of last vertex }
    L, H    : Integer;  { Vertices with lowest & highest F values }
    I, J    : Integer;  { Loop variables }
    Iter    : Integer;  { Iteration number }
    Corr,
    MaxCorr : Float;    { Correcting terms }
    Sum     : Float;
    Flag    : Boolean;

    procedure UpdateSimplex(Y : Float; Q : PVector);
    { Update "worst" vertex and function value }
    begin
      F^[H] := Y;
      CopyVector(P^[H], Q, Lbound, Ubound);
    end;

  begin
    N := Ubound - Lbound + 1;
    M := Succ(Ubound);

    DimMatrix(P, M, Ubound);
    DimVector(F, M);
    DimVector(Pbar, Ubound);
    DimVector(Pstar, Ubound);
    DimVector(P2star, Ubound);

    Iter := 1;
    F0 := MAXNUM;

    { Construct initial simplex }
    for I := Lbound to M do
      CopyVector(P^[I], X, Lbound, Ubound);
    for I := Lbound to Ubound do
      P^[I]^[I] := P^[I]^[I] * Delta;

    { Evaluate function at each vertex }
    for I := Lbound to M do
      F^[I] := ObjFunc(P^[I]);

    repeat
      { Find vertices (L,H) having the lowest and highest
        function values, i.e. "best" and "worst" vertices }
      L := Lbound;
      H := Lbound;
      for I := Succ(Lbound) to M do
        if F^[I] < F^[L] then
          L := I
        else if F^[I] > F^[H] then
          H := I;
      if F^[L] < F0 then
        F0 := F^[L];

      { Find centroid of points other than P(H) }
      for J := Lbound to Ubound do
        begin
          Sum := 0.0;
          for I := Lbound to M do
            if I <> H then Sum := Sum + P^[I]^[J];
          Pbar^[J] := Sum / N;
        end;

      { Reflect worst vertex through centroid }
      for J := Lbound to Ubound do
        Pstar^[J] := 2.0 * Pbar^[J] - P^[H]^[J];
      Ystar := ObjFunc(Pstar);

      { If reflection successful, try extension }
      if Ystar < F^[L] then
        begin
          for J := Lbound to Ubound do
            P2star^[J] := 3.0 * Pstar^[J] - 2.0 * Pbar^[J];
          Y2star := ObjFunc(P2star);
          { Retain extension or contraction }
          if Y2star < F^[L] then
            UpdateSimplex(Y2star, P2star)
          else
            UpdateSimplex(Ystar, Pstar);
        end
      else
        begin
          I := Lbound;
          Flag := False;
          repeat
            if (I <> H) and (F^[I] > Ystar) then Flag := True;
            Inc(I);
          until Flag or (I > M);
          if Flag then
            UpdateSimplex(Ystar, Pstar)
          else
            begin
              { Contraction on the reflection side of the centroid }
              if Ystar <= F^[H] then
                UpdateSimplex(Ystar, Pstar);
              { Contraction on the opposite side of the centroid }
              for J := Lbound to Ubound do
                P2star^[J] := 0.5 * (P^[H]^[J] + Pbar^[J]);
              Y2star := ObjFunc(P2star);
              if Y2star <= F^[H] then
                UpdateSimplex(Y2star, P2star)
              else
                { Contract whole simplex }
                for I := Lbound to M do
                  for J := Lbound to Ubound do
                    P^[I]^[J] := 0.5 * (P^[I]^[J] + P^[L]^[J]);
            end;
        end;

      { Test convergence }
      MaxCorr := 0.0;
      for J := Lbound to Ubound do
        begin
          Corr := Abs(P^[H]^[J] - P^[L]^[J]);
          if Corr > MaxCorr then MaxCorr := Corr;
        end;
      Inc(Iter);
    until (MaxCorr < Tol) or (Iter > MaxIter);

    CopyVector(X, P^[L], Lbound, Ubound);
    F_min := F^[L];

    DelMatrix(P, M, Ubound);
    DelVector(F, M);
    DelVector(Pbar, Ubound);
    DelVector(Pstar, Ubound);
    DelVector(P2star, Ubound);

    if Iter > MaxIter then
      Simplex := NON_CONV
    else
      Simplex := MAT_OK;
  end;

begin
  HessProcAddr := @HessGrad;
end.
