{ **********************************************************************
  *                          Unit REGRESS.PAS                          *
  *                            Version 1.4                             *
  *                      (c) J. Debord, June 1997                      *
  **********************************************************************
                            Regression routines
  ********************************************************************** }

unit Regress;

interface

uses
  FMath, Matrices, Stat;

type
  TRegTest = 
    record           { Test of regression }
      Vr,            { Residual variance }
      R2,            { Coefficient of determination }
      R2a,           { Adjusted coeff. of determination }
      F,             { Variance ratio (explained/residual) }
      Prob : Float;  { Probability of F }
    end;

{ ******************** Unweighted regression routines ******************

  These routines fit equations to data by minimizing the sum of squared
  residuals :
                            n
                      SS = Sum [y(k) - ycalc(k)]^2
                           k=1

  where n is the number of observations, y(k) and ycalc(k) respectively the
  observed and calculated value of the dependent variable for observation k
  ycalc(k) is a function of the regression parameters b(0), b(1) ...

  The following regression types are implemented :

  * simple linear regression :

                        y(k) = b(0) + b(1) * x(k)

  * multiple linear regression :

    y(k) = b(0) + b(1) * x(1,k) + b(2) * x(2,k) + ... + b(Nvar) * x(Nvar,k)

  * polynomial regression :

      y(k) = b(0) + b(1) * x(k) + b(2) * x(k)^2 + ... + b(Deg) * x(k)^Deg

  The parameters passed to/from the regression routines are among the
  following :

  Input  : X        = Vector or matrix of independent variables
           Y        = Vector of dependent variable
           N        = Number of observations
           Nvar     = Number of independent variables
           Deg      = Degree of regression polynomial
           ConsTerm = Flags the presence of a constant term b(0)

  Output : B        = Regression parameters
           V        = Inverse normal equations matrix
  ********************************************************************** }

function LinFit(X, Y : PVector; N : Integer;
                B : PVector; V : PMatrix) : Integer;
{ ----------------------------------------------------------------------
  Simple linear regression
  ---------------------------------------------------------------------- }

function MulFit(X : PMatrix; Y : PVector; N, Nvar : Integer;
                ConsTerm : Boolean; B : PVector; V : PMatrix) : Integer;
{ ----------------------------------------------------------------------
  Multiple linear regression
  ---------------------------------------------------------------------- }

function PolFit(X, Y : PVector; N, Deg : Integer;
                B : PVector; V : PMatrix) : Integer;
{ ----------------------------------------------------------------------
  Polynomial regression
  ---------------------------------------------------------------------- }

{ ******************** Weighted regression routines ********************

  These routines fit equations to data by minimizing the sum of weighted
  squared residuals :
                             n
                      SWS = Sum w(k)*[y(k) - ycalc(k)]^2
                            k=1

  where the "weight" w(k) is inversely proportional to the variance v(k)
  of the observation y(k). v(k) is usually computed as :

                       v(k) = Vr * g[y(k)] = Vr / w(k)

  where Vr is the residual variance and g is a user-specified function
  (e.g. g[y(k)] = y(k)^2 for a constant coefficient of variation).

  Function syntax and results are the same than for unweighted regression
  except that the vector of weights (W) is passed as an additional input
  parameter.
  ********************************************************************** }

function WLinFit(X, Y, W : PVector; N : Integer;
                 B : PVector; V : PMatrix) : Integer;

function WMulFit(X : PMatrix; Y, W : PVector; N, Nvar : Integer;
                 ConsTerm : Boolean; B : PVector; V : PMatrix) : Integer;

function WPolFit(X, Y, W : PVector; N, Deg : Integer;
                 B : PVector; V : PMatrix) : Integer;

{ *********** Routines to test the quality of the regression ***********

  These routines compute the variance-covariance matrix of the fitted 
  parameters and the different statistics used to test the quality of 
  the fit.

  Input parameters  : Y      = Vector of dependent variable
                      Ycalc  = Computed Y values
                      W      = Vector of weights (if any)
                      N      = Number of observations
                      Lbound,
                      Ubound = Indices of first & last fitted parameters
                      V      = Inverse normal equations matrix

  Output parameters : V      = Variance-covariance matrix
                      Test   = Test statistics (Vr, R2, R2a, F, Prob)
  ********************************************************************** }

procedure RegTest(Y, Ycalc : PVector; N, Lbound, Ubound : Integer;
                  V : PMatrix; var Test : TRegTest);
{ ----------------------------------------------------------------------
  Test of unweighted regression
  ---------------------------------------------------------------------- }

procedure WRegTest(Y, Ycalc, W : PVector; N, Lbound, Ubound : Integer;
                   V : PMatrix; var Test : TRegTest);
{ ----------------------------------------------------------------------
  Test of weighted regression
  ---------------------------------------------------------------------- }

{ ****************** Test of regression parameters ********************* }

procedure ParamTest(B : PVector; V : PMatrix; N, Lbound, Ubound : Integer;
                    S, T, Prob : PVector);
{ ----------------------------------------------------------------------
  This routine tests the significance of the parameters. It must be
  called AFTER RegTest or WRegTest since it uses the variance-covariance
  matrix.
  ----------------------------------------------------------------------
  Input parameters  : B      = Regression parameters
                      V      = Variance-covariance matrix
                      N      = Number of observations
                      Lbound,
                      Ubound = Indices of first & last fitted parameters
  ----------------------------------------------------------------------
  Output parameters : S      = Standard deviations of parameters
                      T      = Student's t
                      Prob   = Probabilities
  ---------------------------------------------------------------------- }


implementation

{ Regression modes }
const
  UNWEIGHTED = 0;
  WEIGHTED   = 1;

  function GenLinFit(Mode : Integer; X, Y, W : PVector; N : Integer;
                     B : PVector; V : PMatrix) : Integer;
{ ----------------------------------------------------------------------
  General linear regression routine
  ---------------------------------------------------------------------- }
  var
    WX, S, SX, SY, SX2, SXY, D : Float;
    K : Integer;
  begin
    S := 0.0;
    SX := 0.0;
    SY := 0.0;
    SX2 := 0.0;
    SXY := 0.0;
    if Mode = UNWEIGHTED then
      begin
        S := N;
        for K := 1 to N do
          begin
            SX := SX + X^[K];
            SY := SY + Y^[K];
            SX2 := SX2 + Sqr(X^[K]);
            SXY := SXY + X^[K] * Y^[K];
          end;
      end
    else
      begin
        for K := 1 to N do
          begin
            WX := W^[K] * X^[K];
            S := S + W^[K];
            SX := SX + WX;
            SY := SY + W^[K] * Y^[K];
            SX2 := SX2 + WX * X^[K];
            SXY := SXY + WX * Y^[K];
          end;
      end;
    D := S * SX2 - Sqr(SX);
    if D <= 0.0 then
      GenLinFit := MAT_SINGUL
    else
      begin
        V^[0]^[0] := SX2 / D;
        V^[0]^[1] := - SX / D;
        V^[1]^[0] := V^[0]^[1];
        V^[1]^[1] := S / D;
        B^[0] := V^[0]^[0] * SY + V^[0]^[1] * SXY;
        B^[1] := V^[1]^[0] * SY + V^[1]^[1] * SXY;
        GenLinFit := MAT_OK;
      end;
  end;

  function LinFit(X, Y : PVector; N : Integer;
                  B : PVector; V : PMatrix) : Integer;
  var
    W : PVector;
  begin
    LinFit := GenLinFit(UNWEIGHTED, X, Y, W, N, B, V);
  end;

  function WLinFit(X, Y, W : PVector; N : Integer;
                   B : PVector; V : PMatrix) : Integer;
  begin
    WLinFit := GenLinFit(WEIGHTED, X, Y, W, N, B, V);
  end;

  procedure RegSys(Mode : Integer; X : PMatrix; Y, W : PVector;
                   N, Lbound, Ubound : Integer; A : PMatrix; G : PVector);
{ ----------------------------------------------------------------------
  Sets the system of equations (A.B = G) for multiple linear regression
    y(k) = b(Lbound) * x(Lbound,k) + ... + b(Ubound) * x(Ubound,k)
  ---------------------------------------------------------------------- }
  var
    I, J, K : Integer;
    WX      : Float;
  begin
    { Initialize arrays }
    for I := Lbound to Ubound do
      begin
        for J := I to Ubound do
          A^[I]^[J] := 0.0;
        G^[I] := 0.0;
      end;

    { Compute system matrix and constant vector }
    if Mode = UNWEIGHTED then
      for K := 1 to N do
        for I := Lbound to Ubound do
          begin
            for J := I to Ubound do
              A^[I]^[J] := A^[I]^[J] + X^[I]^[K] * X^[J]^[K];
            G^[I] := G^[I] + X^[I]^[K] * Y^[K];
          end
    else
      for K := 1 to N do
        for I := Lbound to Ubound do
          begin
            WX := W^[K] * X^[I]^[K];
            for J := I to Ubound do
              A^[I]^[J] := A^[I]^[J] + WX * X^[J]^[K];
            G^[I] := G^[I] + WX * Y^[K];
          end;

    { Fill in symmetric matrix }
    for I := Succ(Lbound) to Ubound do
      for J := Lbound to Pred(I) do
        A^[I]^[J] := A^[J]^[I];
  end;

  function GenMulFit(Mode : Integer; X : PMatrix; Y, W : PVector;
                     N, Nvar : Integer; ConsTerm : Boolean;
                     B : PVector; V : PMatrix) : Integer;
{ ----------------------------------------------------------------------
  General multiple linear regression routine
  ---------------------------------------------------------------------- }
  var
    J, K : Integer;
    A    : PMatrix;
    G    : PVector;
  begin
    DimMatrix(A, Nvar, Nvar);
    DimVector(G, Nvar);
    if ConsTerm then
      begin
        if Mode = UNWEIGHTED then
          begin
            A^[0]^[0] := Int(N);
            for K := 1 to N do
              begin
                for J := 1 to Nvar do
                  A^[0]^[J] := A^[0]^[J] + X^[J]^[K];
                G^[0] := G^[0] + Y^[K];
              end;
          end
        else
          begin
            for K := 1 to N do
              begin
                A^[0]^[0] := A^[0]^[0] + W^[K];
                for J := 1 to Nvar do
                  A^[0]^[J] := A^[0]^[J] + W^[K] * X^[J]^[K];
                G^[0] := G^[0] + W^[K] * Y^[K];
              end;
          end;
        for J := 1 to Nvar do
          A^[J]^[0] := A^[0]^[J];
      end;
    RegSys(Mode, X, Y, W, N, 1, Nvar, A, G);
    if ConsTerm then
      GenMulFit := GaussJordan(A, G, 0, Nvar, V, B)
    else
      GenMulFit := GaussJordan(A, G, 1, Nvar, V, B);
    DelMatrix(A, Nvar, Nvar);
    DelVector(G, Nvar);
  end;

  function MulFit(X : PMatrix; Y : PVector; N, Nvar : Integer;
                  ConsTerm : Boolean; B : PVector; V : PMatrix) : Integer;
  var
    W : PVector;
  begin
    MulFit := GenMulFit(UNWEIGHTED, X, Y, W, N, Nvar, ConsTerm, B, V);
  end;

  function WMulFit(X : PMatrix; Y, W : PVector; N, Nvar : Integer;
                   ConsTerm : Boolean; B : PVector; V : PMatrix) : Integer;
  begin
    WMulFit := GenMulFit(WEIGHTED, X, Y, W, N, Nvar, ConsTerm, B, V);
  end;

  procedure PowMat(X : PVector; N, Deg : Integer; U : PMatrix);
{ ----------------------------------------------------------------------
  Computes matrix of increasing powers of X for polynomial regression
  ---------------------------------------------------------------------- }
  var
    I, K : Integer;
  begin
    for K := 1 to N do
      begin
        U^[1]^[K] := X^[K];
        for I := 2 to Deg do
          U^[I]^[K] := U^[I - 1]^[K] * X^[K];
      end;
  end;

  function GenPolFit(Mode : Integer; X, Y, W : PVector; N, Deg : Integer;
                     B : PVector; V : PMatrix) : Integer;
{ ----------------------------------------------------------------------
  General polynomial regression routine
  ---------------------------------------------------------------------- }
  var
    U : PMatrix;
  begin
    DimMatrix(U, Deg, N);
    PowMat(X, N, Deg, U);
    GenPolFit := GenMulFit(Mode, U, Y, W, N, Deg, True, B, V);
    DelMatrix(U, Deg, N);
  end;

  function PolFit(X, Y : PVector; N, Deg : Integer;
                  B : PVector; V : PMatrix) : Integer;
  var
    W : PVector;
  begin
    PolFit := GenPolFit(UNWEIGHTED, X, Y, W, N, Deg, B, V);
  end;

  function WPolFit(X, Y, W : PVector; N, Deg : Integer;
                   B : PVector; V : PMatrix) : Integer;
  begin
    WPolFit := GenPolFit(WEIGHTED, X, Y, W, N, Deg, B, V);
  end;

  procedure GenRegTest(Mode : Integer; Y, Ycalc, W : PVector;
                       N, Lbound, Ubound : Integer; V : PMatrix;
                       var Test : TRegTest);
  var
    Ybar     : Float;    { Average Y value }
    SSt      : Float;    { Total sum of squares }
    SSe      : Float;    { Explained sum of squares }
    SSr      : Float;    { Residual sum of squares }
    Npar     : Integer;  { Number of fitted parameters }
    Nu1, Nu2 : Integer;  { Degrees of freedom }
    I, J     : Integer;  { Loop variables }
  begin
    Npar := Ubound - Lbound + 1;
    with Test do
      if N > Npar then
        begin
          Ybar := Average(Y, 1, N);
          if Mode = UNWEIGHTED then
            begin
              SSt := SumSqrDif(Y, 1, N, Ybar);
              SSe := SumSqrDif(Ycalc, 1, N, Ybar);
              SSr := SumSqrDifVect(Y, Ycalc, 1, N);
            end
          else
            begin
              SSt := SumWSqrDif(Y, W, 1, N, Ybar);
              SSe := SumWSqrDif(Ycalc, W, 1, N, Ybar);
              SSr := SumWSqrDifVect(Y, Ycalc, W, 1, N);
            end;
          Nu1 := Npar - 1;
          Nu2 := N - Npar;
          R2 := SSe / SSt;
          R2a := 1.0 - (1.0 - R2) * (N - 1) / Nu2;
          Vr := SSr / Nu2;
          if Vr > 0.0 then
            begin
              F := (SSe / Nu1) / Vr;
              Prob := PSnedecor(Nu1, Nu2, F);
            end
          else
            begin
              F := MAXNUM;
              Prob := 0.0;
            end;
        end
      else
        begin
          Vr := 0.0;
          R2 := 1.0;
          R2a := 0.0;
          F := 0.0;
          Prob := 1.0;
        end;
    { Compute variance-covariance matrix }
    for I := Lbound to Ubound do
      for J := I to Ubound do
        V^[I]^[J] := V^[I]^[J] * Test.Vr;
    for I := Succ(Lbound) to Ubound do
      for J := Lbound to Pred(I) do
        V^[I]^[J] := V^[J]^[I];
  end;

  procedure RegTest(Y, Ycalc : PVector; N, Lbound, Ubound : Integer;
                    V : PMatrix; var Test : TRegTest);
  var
    W : PVector;
  begin
    GenRegTest(UNWEIGHTED, Y, Ycalc, W, N, Lbound, Ubound, V, Test);
  end;

  procedure WRegTest(Y, Ycalc, W : PVector; N, Lbound, Ubound : Integer;
                     V : PMatrix; var Test : TRegTest);
  begin
    GenRegTest(WEIGHTED, Y, Ycalc, W, N, Lbound, Ubound, V, Test);
  end;

  procedure ParamTest(B : PVector; V : PMatrix; N, Lbound, Ubound : Integer;
                      S, T, Prob : PVector);
  var
    I  : Integer;
    Nu : Integer;  { Degrees of freedom }
  begin
    Nu := N - (Ubound - Lbound + 1);  { DoF = Nb points - Nb parameters }
    for I := Lbound to Ubound do
      begin
        S^[I] := Sqrt(V^[I]^[I]);
        T^[I] := B^[I] / S^[I];
        Prob^[I] := PStudent(Nu, T^[I]);
      end;
  end;

end.
